
EBNF-like grammar for simple regular expressions (SREs)
;******************************************************

sre	   = term {"|" term}.

term	   = factor {factor}.

factor	   = atom [un_op].

atom	   = "(" sre ")"
	   | char_set
	   | char
	   | ".".	; interpreted as [^\x0A\x0D]
			; i.e. all but CR and LF (Windows newline sequence)

un_op	   = "*"|"+"|"?".

char_set   = "[" char_list ["-"] "]". ; if "-" belongs to the set it must
				      ; appear in the last position

char_list  = "^"  list	   ; set complement operator (must appear first)
	   | "]"  list	   ; if "]" belongs to the set it must appear first
	   | "^]" list.    ; or directly after the complement operator
	   | list.

list	   = char ["-" char] {char ["-" char]}.

char	   = PRINT
	   | esc_seq.

esc_seq    = "\" (oct_esq|hex_esq|c_esq|spc_esc|PRINT).

oct_esc    = ODIGIT [ODIGIT [ODIGIT]].

hex_esc    = ("x"|"X") XDIGIT [XDIGIT].

c_esc	   = "a"|"b"|"f"|"n"|"r"|"t"|"v".

spc_esc    = "|"|"("|")"|"*"|"+"|"?"|"."|"["|"]".

PRINT	   = GRAPH
	   | " ".

GRAPH	   = ALNUM
	   | PUNCT.

ALNUM	   = ALPHA
	   | DIGIT.

ALPHA	   = LOWER
	   | UPPER.

PUNCT	   = "!"|'"'|"#"|"%"|"&"|"'"|"("|")"|";"
	   | "<"|"="|">"|"?"|"["|"\"|"]"|"*"|"+"
	   | ","|"-"|"."|"/"|":"|"^"|"_"|"{"|"|"
	   | "}"|"~"|"`"|"@".

DIGIT	   = "0"|"1"|...|"9".

LOWER	   = "a"|"b"|...|"z".

UPPER	   = "A"|"B"|...|"Z".

CNTRL	   = 0|1|...|31|127.

BLANK	   = 9|" ".

SPACE	   = 9|10|11|12|13|" ".

XDIGIT	   = "a"|"b"|...|"f"
	   | "A"|"B"|...|"F"
	   | DIGIT.

ODIGIT	   = "0"|"1"|...|"7".
