
CR  equ 0Dh
LF  equ 0Ah
HT  equ 09h

C_OPTION   equ	1
N_OPTION   equ	2
Q_OPTION   equ	4
V_OPTION   equ	8


MAX_ARGV  equ	 4	; maximal number of command line arguments
			; (without program name and switches, i.e.,
			;  only "real" arguments)
MAX_ARGS  equ  512	; size of buffer for 0 terminated argument strings

; DATA (copy to .bss section and uncomment)
;
; argc  rd 1
; argv  rd MAX_ARGV
; args  rd MAX_ARGS

;******************************************************************************
;
; Parse command line and setup command line arguments.
;
; IN  : esi    == pointer to command line string
;
; OUT : eax    == 0 if error
;       ecx    == pointer to error message
;       edx    == length  of error message
;       -- no error --
;       eax    != 0 if ok
;       ebx    == pointer to next free argv element
;       edi    == pointer to next free args byte
;       [argc] == number of command line arguments (without program name)
;       [argv] == array of [argc] pointers to command line arguments
;       [args] == array of 0 terminated strings pointed to by [argv] elements
;
;******************************************************************************
parse_cmdline:
	xor	eax, eax
	mov	ebx, argv
	mov	edi, args
	mov	ecx, MAX_ARGS
	mov	edx, MAX_ARGV
	mov	[argc], eax

   .skip_whitespaces:
	lodsb
	cmp	eax, " "
	jz   .skip_whitespaces
	cmp	eax, HT
	jz   .skip_whitespaces
	cmp	eax, '"'
	jz   .find_quote

   .skip_prog_name:
	lodsb
	test	eax, 80h
	jnz  .skip_prog_name
	test	byte[ctype_map+eax], BLANK
	jnz  .scan_args
	test	byte[ctype_map+eax], CNTRL
	jz   .skip_prog_name
	jmp  .no_args
	;--------------------

   .scan_args:
	lodsb
	test	eax, eax
	jz   .finished
	test	eax, 80h
	jnz  .init_arg
	test	byte[ctype_map+eax], BLANK
	jnz  .scan_args
	test	byte[ctype_map+eax], CNTRL
	jnz  .finished
	cmp	eax, "-"
	jz   .get_option

   .init_arg:
	inc	dword[argc]
	dec	edx
	jl   .too_many_args
	mov	[ebx], edi
	cmp	eax, '"'
	lea	ebx, [ebx+4]
	jz   .quoted_param
	dec	ecx
	jz   .args_too_long
	stosb

   .copy_param:
	lodsb
	test	eax, 80h
	jnz  .copy_char
	test	byte[ctype_map+eax], BLANK
	jnz  .next_param
	test	byte[ctype_map+eax], CNTRL
	jnz  .last_param

   .copy_char:
	dec	ecx
	jz   .args_too_long
	stosb
	jmp  .copy_param
	;--------------------

   .next_param:
	mov	byte[edi],0
	inc	edi
	jmp  .scan_args
	;--------------------

   .quoted_param:
	lodsb
	test	eax, eax
	jz   .last_param
	cmp	eax, CR
	jz   .last_param
	cmp	eax, LF
	jz   .last_param
	cmp	eax, '"'
	jz   .next_param
	dec	ecx
	jz   .args_too_long
	stosb
	jmp  .quoted_param
	;--------------------

   .get_option:
	push	ebx

    .find_option:
	lodsb
	xor	ebx, ebx
	test	byte[ctype_map+eax], BLANK
	jnz  .find_option	 ; allow " " or \t between "-" and option char
	test	byte[ctype_map+eax], CNTRL
	jnz  .finished
	inc	ebx
	cmp	eax, "c"	; display only match count
	jz   .set_option
	shl	ebx, 1
	cmp	eax, "n"	; display line numbers of match lines
	jz   .set_option
	shl	ebx, 1
	cmp	eax, "q"	; set return code only
	jz   .set_option
	shl	ebx, 1
	cmp	eax, "v"	; display non matching lines
	jnz  .invalid_option	; invalid option --> print usage message

   .set_option:
	or	dword[cmdln_options], ebx
	movzx	eax, byte[esi]
	pop	ebx
	test	eax, 80h
	jnz  .invalid_arg
	test	byte[ctype_map+eax], LOWER
	; allow multiple options, e.g., "-cv"
	jnz  .get_option
	jmp  .scan_args
	;--------------------

   .find_quote:
	lodsb
	cmp	eax, HT
	jz   .find_quote
	test	byte[ctype_map+eax], CNTRL
	jnz  .broken_string
	cmp	eax, '"'
	jnz  .find_quote
	jmp  .scan_args
	;--------------------

   .last_param:
	xor	eax, eax
	dec	ecx
	jz   .args_too_long
	stosb

   .finished:
	xor	eax, eax
	cmp	[argc], eax
	jz   .no_args
	inc	eax
	retn
	;====================

   .args_too_long:
	mov	ecx, err_cmdln_too_long_str
	mov	edx, err_cmdln_too_long_len
	jmp  .error
	;--------------------

   .too_many_args:
	mov	ecx, err_too_many_args_str
	mov	edx, err_too_many_args_len
	jmp  .error
	;--------------------

   .invalid_option:
	pop	ebx
	mov	ecx, err_inv_option_str
	mov	edx, err_inv_option_len
	jmp  .error
	;--------------------

   .invalid_arg:
	mov	ecx, err_inv_arg_str
	mov	edx, err_inv_arg_len
	jmp  .error
	;--------------------

   .broken_string:
	mov	ecx, err_broken_string_str
	mov	edx, err_broken_string_len
	jmp  .error
	;--------------------

   .no_args:
	mov	ecx, usage_str
	mov	edx, usage_len

   .error:
	xor	eax, eax
	retn
	;====================

;endp parse_cmdline
