
; System call wrappers for Windows
;---------------------------------

; file related WIN32 constants

OPEN_EXISTING		equ 03h
FILE_SHARE_READ 	equ 01h
GENERIC_READ		equ 80000000h

MAX_PATH		equ 260
INVALID_HANDLE_VALUE	equ -1

FILE_ATTRIBUTE_READONLY 		equ    1h
FILE_ATTRIBUTE_HIDDEN			equ    2h
FILE_ATTRIBUTE_SYSTEM			equ    4h
FILE_ATTRIBUTE_DIRECTORY		equ   10h
FILE_ATTRIBUTE_ARCHIVE			equ   20h
FILE_ATTRIBUTE_DEVICE			equ   40h
FILE_ATTRIBUTE_NORMAL			equ   80h
FILE_ATTRIBUTE_TEMPORARY		equ  100h
FILE_ATTRIBUTE_SPARSE_FILE		equ  200h
FILE_ATTRIBUTE_REPARSE_POINT		equ  400h
FILE_ATTRIBUTE_COMPRESSED		equ  800h
FILE_ATTRIBUTE_OFFLINE			equ 1000h
FILE_ATTRIBUTE_NOT_CONTENT_INDEXED	equ 2000h
FILE_ATTRIBUTE_ENCRYPTED		equ 4000h

struc FILETIME
{
    .dwLowDateTime	rd 1
    .dwHighDateTime	rd 1
}
SIZE_FILETIME		equ 8

dwLowDateTime@FILETIME	equ 0
dwHighDateTime@FILETIME equ 4

struc WIN32_FIND_DATA
{
    .dwFileAttributes	rd 1
    .ftCreationTime	FILETIME
    .ftLastAccessTime	FILETIME
    .ftLastWriteTime	FILETIME
    .nFileSizeHigh	rd 1
    .nFileSizeLow	rd 1
    .dwReserved0	rd 1
    .dwReserved1	rd 1
    .cFileName		rb MAX_PATH
    .cAlternateFileName rb 14
}
SIZE_WIN32_FIND_DATA	equ ((5 shl 2) + (3 * SIZE_FILETIME) + MAX_PATH + 14)

dwFileAttributes@WIN32_FIND_DATA   equ	0
ftCreationTime@WIN32_FIND_DATA	   equ	4
ftLastAccessTime@WIN32_FIND_DATA   equ	(4+SIZE_FILETIME)
ftLastWriteTime@WIN32_FIND_DATA    equ	(ftLastAccessTime@WIN32_FIND_DATA + SIZE_FILETIME)
nFileSizeHigh@WIN32_FIND_DATA	   equ	(ftLastWriteTime@WIN32_FIND_DATA + SIZE_FILETIME)
nFileSizeLow@WIN32_FIND_DATA	   equ	(nFileSizeHigh@WIN32_FIND_DATA+4)
dwReserved0@WIN32_FIND_DATA	   equ	(nFileSizeLow@WIN32_FIND_DATA+4)
dwReserved1@WIN32_FIND_DATA	   equ	(dwReserved0@WIN32_FIND_DATA+4)
cFileName@WIN32_FIND_DATA	   equ	(dwReserved1@WIN32_FIND_DATA+4)
cAlternateFileName@WIN32_FIND_DATA equ	(cFileName@WIN32_FIND_DATA+MAX_PATH)


; console related WIN32 constants

STD_INPUT_HANDLE	 equ -10
STD_OUTPUT_HANDLE	 equ -11
STD_ERROR_HANDLE	 equ -12


; IN  : nothing
; OUT : eax == pointer to the command line string
; REGS: ecx, edx may be destroyed
macro	@sys.get_commandline
{
	call	[GetCommandLineA]
}

; IN  : std_handle == STD_INPUT_HANDLE or STD_OUTPUT_HANDLE or STD_ERROR_HANDLE
; OUT : eax == INVALID_HANDLE_VALUE if error or the specified handle if success
; REGS: ecx, edx may be destroyed
macro	@sys.get_stdhandle  std_handle
{
	push	std_handle
	call	[GetStdHandle]
}

; file:
;----------
; fopen
; IN  :ebx == pointer to zero terminated file name string
; OUT :eax == file handle if success or INVALID_HANDLE_VALUE if error
; REGS:ecx, edx may be destroyed
macro  @sys.fopen
{
	push	0
	push	FILE_ATTRIBUTE_NORMAL
	push	OPEN_EXISTING
	push	0
	push	FILE_SHARE_READ
	push	GENERIC_READ
	push	ebx
	call	[CreateFileA]
}

; fsize
; IN  :eax == file handle
; OUT :eax == file size + 1
; REGS:ecx, edx may be destroyed
macro  @fsize
{
	push	0
	push	eax
	call	[GetFileSize]
	inc	eax		 ; + EOF
}

; fread
; IN  :ebx == file handle
;      ecx == pointer to buffer for file contents
;      edx == number  of bytes to read from file
; OUT :eax == 0 if error, != 0 if success
;      ecx == number  of bytes actually read from file handle (if eax != 0)
; REGS:ecx, edx may be destroyed
macro  @sys.fread
{
	push	0
	mov	esi, esp

	push	0
	push	esi
	push	edx
	push	ecx
	push	ebx
	call	[ReadFile]
	pop	ecx
}

; fwrite
; IN  :ebx == file handle
;      ecx == pointer to buffer to write to file
;      edx == number  of bytes to write to file
; OUT :eax == 0 if error, != 0 if success
;      ecx == number  of bytes actually written to file handle (if eax != 0)
; REGS:ecx, edx may be destroyed
macro  @sys.fwrite
{
	push	0	 ; bytes_rw
	mov	eax, esp

	push	0
	push	eax
	push	edx
	push	ecx
	push	ebx
	call	[WriteFile]
	pop	ecx	 ; bytes_rw
}

; fclose
; IN   :ebx == file handle
; OUT  :eax == 0 if error, != 0 if success
; REGS :ecx, edx may be destroyed
macro  @sys.fclose
{
	push	ebx
	call	[CloseHandle]
}

; exit
; IN   : ebx == return code
; OUT  : nothing
macro  @sys.exit
{
	push	ebx
	call	[ExitProcess]
}

; console:
;-------------
; print_string
; PRE :stdout_handle must be properly initialized
; IN  :ecx == pointer to string
;      edx == length of string
; OUT :eax == 0 if error, =! 0 if success
;      ecx == number of bytes written to stdout_handle (if eax != 0)
; REGS:edx may be destroyed

IF USED sys.print_string

sys.print_string:
	mov	ebx, [stdout_handle]
       @sys.fwrite
	retn

END IF
